# encoding: utf-8
from functools import wraps
from carbon.client import stat, metrics


def timeit(name, client=stat):
    def decorator(func):
        @wraps(func)
        def wrap(*args, **kwargs):
            client[name] = metrics.Timer
            watch = client[name].start()
            res = func(*args, **kwargs)
            client[name].stop(watch)
            return res
        return wrap
    return decorator



# encoding: utf-8
from carbon.client.metrics.base import MeasurerBase, Metric


class Collector(MeasurerBase):

    def add(self, value):
        self._values.add(Metric(name=self.name, value=value))



# encoding: utf-8
from functools import total_ordering
from time import time
from abc import abstractmethod


@total_ordering
class Metric(object):
    __slots__ = '__value', '__ts', '__name'

    def __init__(self, name, value, ts=None):
        if not ts:
            ts = time()

        self.__ts = ts
        self.__name = name
        self.__value = value

    def __str__(self):
        return "%s %s %.6f" % (self.__name, self.__value, self.__ts)

    def __hash__(self):
        return hash(tuple(self))

    def __iter__(self):
        yield self.__name
        yield self.__value
        yield self.__ts

    def __eq__(self, other):
        if not isinstance(other, Metric):
            raise ValueError("Can't check equality with non-Metric object")

        return tuple(self) == tuple(other)

    def __gt__(self, other):
        if not isinstance(other, Metric):
            raise ValueError("Can't check equality with non-Metric object")

        return tuple(self) > tuple(other)


class MeasurerAbstract(object):                 #pragma: nocover
    @abstractmethod
    def __init__(self, cleanup=None):
        raise NotImplementedError

    @abstractmethod
    def on_create(self):
        raise NotImplementedError

    @abstractmethod
    def add(self, metric):
        raise NotImplementedError

    @abstractmethod
    def on_delete(self):
        raise NotImplementedError

    @abstractmethod
    def on_send(self):
        raise NotImplementedError

    @abstractmethod
    def str(self, ns):
        raise NotImplementedError


class MeasurerBase(MeasurerAbstract):
    __slots__ = '_value', '_cleanup', '_values', 'name'

    def __init__(self, cleanup=None):
        self.name = None
        self._cleanup = cleanup
        self._values = set([])

    def on_create(self):
        return

    def add(self, metric):
        assert isinstance(metric, Metric), "It's not a Metric"
        self._values.add(metric)

    def on_delete(self):
        if self._cleanup:
            self._cleanup()

    def on_send(self):
        self._values = set([])

    def str(self, ns):
        return "\n".join(map(lambda x: "%s.%s" % (ns, x) if x else '', list(self._values)))


MetricTypeBase = MeasurerBase



import asyncio
import socket
import logging
from .udp import UDPClient as BaseUDPClient


log = logging.getLogger(__name__)


def create_future(loop):
    if hasattr(loop, 'create_future'):
        return loop.create_future()
    else:
        return asyncio.Future(loop=loop)


class AsyncUDPSocket:

    __slots__ = '__loop', '__sock', '__address', '__futures', '__closed', '__writer_added'

    def __init__(self, loop=None):
        self.__loop = asyncio.get_event_loop() if loop is None else loop
        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.__sock.setblocking(False)

        self.__futures = list()
        self.__closed = False
        self.__writer_added = False

    def sendto(self, data, host, port):
        data = data if isinstance(data, bytes) else str(data).encode('utf-8')
        future = create_future(self.__loop)

        destination = (data, host, port)
        self.__futures.append((destination, future))

        if not self.__writer_added:
            self.__loop.add_writer(self.__sock.fileno(), self.__sender)
            self.__writer_added = True

        return future

    def __sender(self):
        if not self.__futures:
            self.__loop.remove_writer(self.__sock.fileno())
            self.__writer_added = False
            return

        destination, future = self.__futures[0]
        data, host, port = destination

        try:
            self.__sock.sendto(data, (host, port))
        except (BlockingIOError, InterruptedError):
            return
        except Exception as exc:
            self.__abort(exc)
        else:
            self.__futures.pop(0)
            future.set_result(True)

    def __abort(self, exc):
        for future in (f for _, f in self.__futures if not f.done()):
            future.set_exception(exc)

        self.close()

    @property
    def is_closed(self):
        return self.__closed

    def close(self):
        if self.__closed:
            raise RuntimeError("Socket already closed")

        self.__closed = True
        self.__loop.remove_writer(self.__sock.fileno())
        self.__sock.close()

        for future in (f for _, f in self.__futures if not f.done()):
            future.set_exception(ConnectionError("Connection closed"))


class UDPClient(BaseUDPClient):
    def __init__(self, hosts='127.0.0.1', ns='carbon.client', loop=None):
        super().__init__(hosts=hosts, ns=ns)

        self.__loop = loop or asyncio.get_event_loop()
        self.__socket = AsyncUDPSocket(loop=self.__loop)

    @property
    def socket(self):
        return self.__socket

    def _sender(self, packet, host, port):
        asyncio.ensure_future(self.socket.sendto(packet, host, port), loop=self.__loop)



# encoding: utf-8
from carbon.client.udp import UDPClient
from carbon.client.metrics import Counter, Timer, Collector
from carbon.client import stat


class SimpleMeasurerBase(object):
    __slots__ = '_client', '_metric'

    def __init__(self, metric, client=None):
        if client is None:
            client = stat

        assert isinstance(client, UDPClient)
        self._client = client
        self._metric = metric


class SimpleCounter(SimpleMeasurerBase):
    TYPE = Counter

    def inc(self, status):
        metric = "{}_{}".format(self._metric, status)
        self._client[metric] = self.TYPE
        self._client[metric].inc(1)

    def __enter__(self):
        self.inc("ok")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            self.inc("fail")


class SimpleTimer(SimpleMeasurerBase):
    TYPE = Timer

    def __enter__(self):
        self._done_metric = "%s_ok" % self._metric
        self._fail_metric = "%s_fail" % self._metric

        self._client[self._done_metric] = self.TYPE
        self._client[self._fail_metric] = self.TYPE

        self._done_watch = self._client[self._done_metric].start()
        self._fail_watch = self._client[self._fail_metric].start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            metric = self._done_metric
            watch = self._done_watch
        else:
            metric = self._fail_metric
            watch = self._fail_watch

        self._client[metric].stop(watch)


class SimpleCollector(SimpleMeasurerBase):
    TYPE = Collector

    def __enter__(self):
        if self._metric not in self._client:
            self._client[self._metric] = self.TYPE

        return self._client[self._metric]

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


SimpleMetricBase = SimpleMeasurerBase



# encoding: utf-8
from time import time
from threading import RLock
from carbon.client.metrics.base import MeasurerBase, Metric


class StopWatch(object):
    __slots__ = '_lock', '_current'

    def __init__(self):
        self._lock = RLock()
        self._current = None

    def start(self):
        with self._lock:
            self._current = time()

    def stop(self):
        assert self._current, "StopWatch not running"
        with self._lock:
            return time() - self._current


class Timer(MeasurerBase):
    __slots__ = '_current',

    def __init__(self, cleanup=None):
        MeasurerBase.__init__(self, cleanup)
        self._current = None

    @classmethod
    def start(cls):
        watch = StopWatch()
        watch.start()
        return watch

    def stop(self, stop_watch):
        assert isinstance(stop_watch, StopWatch)
        self.add(Metric(name=self.name, value=stop_watch.stop()))



# encoding: utf-8
import re
import logging
from functools import wraps
from socket import socket, AF_INET, SOCK_DGRAM

from carbon.client import metrics
from carbon.client.metrics.base import MeasurerBase
from carbon.client.compat import b, basestring


log = logging.getLogger('carbon.client.udp')


class LockFlag(object):

    def __init__(self):
        self.__flag = False

    def __call__(self, func):
        @wraps(func)
        def wrap(*args, **kwargs):
            if self.__flag:
                return

            self.__flag = True
            try:
                func(*args, **kwargs)
            finally:
                self.__flag = False

        return wrap


class UDPClient(object):
    LOCK = LockFlag()
    DEFAULT = metrics.Counter

    __slots__ = '__host', '__ns', '__socket', '__endpoints', '__sending', '__metrics'

    def __init__(self, hosts='127.0.0.1', ns='carbon.client'):
        self.__host = hosts
        self.__ns = ns
        self.__socket = None

        self.__sending = False
        self.__metrics = {}
        self.__endpoints = self._parse_hosts(hosts)
        self.__add_metric(metrics.HeartBeat)

    @staticmethod
    def _parse_hosts(hosts):
        endpoints = list()

        for host_str in hosts.split(","):
            if ":" in host_str:
                host, port = host_str.split(":", 1)
                port = int(port)
            else:
                host, port = host_str, 2003

            if not isinstance(host, str):
                host = host.decode()

            endpoints.append((host, port))

        return frozenset(endpoints)

    @property
    def hosts(self):
        return self.__endpoints

    @hosts.setter
    @LOCK
    def hosts(self, hosts):
        self.__endpoints = self._parse_hosts(hosts)

    @property
    def ns(self):
        return self.__ns

    @ns.setter
    def ns(self, ns):
        if ns.startswith('.'):
            raise ValueError("NameSpace mustn't starts with the dot.")

        if ns.endswith('.'):
            raise ValueError("NameSpace mustn't ends with the dot.")

        if re.match('^[\w\d\._\-]+$', ns) is None:
            raise ValueError("NameSpace mustn't contain special chars (except '_', '-')")

        if bool(list(filter(lambda x: not x, ns.split('.')))):
            raise ValueError("Namespace must contain chars after dots.")

        self.__ns = ns

    def __add_metric(self, metric, name=None):
        metric_instance = metric(lambda x: self.__metrics.remove(x))
        metric_instance.name = name
        metric_instance.on_create()
        self.__metrics[name] = metric_instance

    def __contains__(self, item):
        return item in self.__metrics

    @property
    def socket(self):
        if not self.__socket:
            self.__socket = socket(AF_INET, SOCK_DGRAM)
        return self.__socket

    def __getitem__(self, item):
        if item not in self.__metrics:
            self.__add_metric(self.DEFAULT, item)
        return self.__metrics[item]

    def __setitem__(self, name, metric_type):
        if not issubclass(metric_type, MeasurerBase):
            raise ValueError("Unknown metric type")

        if not isinstance(name, basestring):
            raise ValueError("Invalid metric name")

        if name not in self or self.__metrics.get(name) is not metric_type:
            self.__add_metric(metric_type, name)

    def _sender(self, packet, host, port):
        self.socket.sendto(packet, (host, port))

    @LOCK
    def send(self):
        metric_set = list(self.__metrics.values())

        packet = b(
            "\n".join(
                filter(
                    lambda x: x,
                    map(
                        lambda x: x.str(self.__ns),
                        metric_set
                    )
                )
            )
        )

        if isinstance(packet, basestring):
            packet = packet.encode()

        for host, port in self.__endpoints:
            try:
                self._sender(packet, host, port)
            except Exception as e:
                log.exception("Error when sending data to carbon")

        for m in metric_set:
            m.on_send()

    def close(self):
        self.socket.close()



# encoding: utf-8
from carbon.client.metrics.base import Metric, MeasurerBase


class HeartBeat(MeasurerBase):
    __slots__ = 'value',

    def __init__(self, cleanup=None):
        super(HeartBeat, self).__init__(cleanup)
        self.value = 0
        self.add(Metric(name='heartbeat', value=self.value))

    def on_send(self):
        self.value += 1
        self._values.clear()
        self.add(Metric(name='heartbeat', value=self.value))



# encoding: utf-8
from __future__ import absolute_import
import os

from carbon.client.udp import UDPClient

stat = UDPClient(
    hosts=os.getenv('CARBON_HOSTS', os.getenv('CARBON_HOST', '127.0.0.1')),
    ns=os.getenv(
        'CARBON_NS',
        os.getenv(
            'CARBON_NAMESPACE',
            'carbonate'
        )
    )
)


from . import metrics
from . import decorators
from . import extras


__all__ = ['metrics', 'decorators', 'UDPClient', 'stat']



# encoding: utf-8
from __future__ import absolute_import, print_function
from setuptools import setup, find_packages


__version__ = '0.3.3'
__author__ = 'Dmitry Orlov <me@mosquito.su>'


setup(
    name='carbon-client',
    version=__version__,
    author=__author__,
    author_email='me@mosquito.su',
    license="MIT",
    description="graphite/carbon udp client for sending metrics",
    platforms="all",
    url="http://github.com/mosquito/carbon-client",
    classifiers=[
        'Environment :: Console',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: System :: Monitoring',
        'Topic :: System :: Networking :: Monitoring',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Natural Language :: Russian',
        'Topic :: Software Development :: Libraries',
        'Development Status :: 4 - Beta',

    ],
    long_description=open('README.rst').read(),
    packages=find_packages(exclude=('tests',)),
    install_requires=[],
)



import sys


PY3 = sys.version_info > (3,)


try:
    basestring = basestring
except NameError:
    basestring = str


def b(obj):
    if PY3:
        if isinstance(obj, str):
            return obj.encode('utf-8')
    else:
        if not isinstance(obj, str):
            return obj.encode('utf-8')

    return obj



# encoding: utf-8
from carbon.client.metrics.base import Metric, MetricTypeBase


class Counter(MetricTypeBase):

    def __init__(self, cleanup=None):
        MetricTypeBase.__init__(self, cleanup)
        self.value = 0

    def add(self, value):
        self.value += value
        self._values.add(Metric(name=self.name, value=self.value))

    def inc(self, n=1):
        return self.add(n)

    def dec(self, n=1):
        return self.add(-n)
